# coding=utf-8
"""Module taken from OpenStudio-standards.

Prototype fan calculation methods that are the same regardless of fan type.
These methods are available to FanConstantVolume, FanOnOff, FanVariableVolume,
and FanZoneExhaust.

https://github.com/NREL/openstudio-standards/blob/master/
lib/openstudio-standards/prototypes/common/objects/Prototype.Fan.rb
"""
from __future__ import division
import os
import json

from ladybug.datatype.pressure import Pressure

from honeybee_openstudio.openstudio import openstudio_model


def _load_json_data(json_name):
    """load the fan data from the JSON file."""
    try:
        cur_dir = os.path.dirname(__file__)
        data_file = os.path.join(cur_dir, 'data', json_name)
        with open(data_file, 'r') as inf:
            fan_data = json.load(inf)
    except IOError:
        fan_data = {}
        print('Failed to import fan JSON data.')
    return fan_data


def _fan_curve_coefficients_from_json(fan_curve_name):
    """Lookup fan curve coefficients."""
    for curve_obj in CURVE_DATA:
        if curve_obj['name'] == fan_curve_name:
            return curve_obj['coeff_1'], curve_obj['coeff_2'], \
                curve_obj['coeff_3'], curve_obj['coeff_4'], curve_obj['coeff_5']
    return None, None, None, None, None


PRESSURE = Pressure()
FAN_DATA = _load_json_data('ashrae_90_1.fans.json')['fans']
CURVE_DATA = _load_json_data('ashrae_90_1.curves.json')['curves']


def create_fan_by_name(
        model, standards_name, fan_name=None, fan_efficiency=None, pressure_rise=None,
        motor_efficiency=None, motor_in_airstream_fraction=None,
        fan_power_minimum_flow_rate_input_method=None,
        fan_power_minimum_flow_rate_fraction=None,
        system_availability_manager_coupling_mode=None,
        end_use_subcategory=None):
    """Create a fan with properties for a fan name in the standards data.

    Args:
        fan_name: [String] fan name.
        fan_efficiency: [Double] fan efficiency.
        pressure_rise: [Double] fan pressure rise in Pa.
        end_use_subcategory: [String] end use subcategory name.
    """
    # get the data from the JSON
    for fan_obj in FAN_DATA:
        if fan_obj['name'] == standards_name:
            fan_json = fan_obj
            break

    # create the fan
    if fan_json['type'] == 'ConstantVolume':
        fan = create_fan_constant_volume_from_json(
            model, fan_json, fan_name=fan_name, fan_efficiency=fan_efficiency,
            pressure_rise=pressure_rise, motor_efficiency=motor_efficiency,
            motor_in_airstream_fraction=motor_in_airstream_fraction,
            end_use_subcategory=end_use_subcategory)
    elif fan_json['type'] == 'OnOff':
        fan = create_fan_on_off_from_json(
            model, fan_json, fan_name=fan_name, fan_efficiency=fan_efficiency,
            pressure_rise=pressure_rise, motor_efficiency=motor_efficiency,
            motor_in_airstream_fraction=motor_in_airstream_fraction,
            end_use_subcategory=end_use_subcategory)
    elif fan_json['type'] == 'VariableVolume':
        fpc_1, fpc_2, fpc_3, fpc_4, fpc_5 = \
            _fan_curve_coefficients_from_json(fan_json['fan_curve'])
        fan = create_fan_variable_volume_from_json(
            model, fan_json, fan_name=fan_name, fan_efficiency=fan_efficiency,
            pressure_rise=pressure_rise, motor_efficiency=motor_efficiency,
            motor_in_airstream_fraction=motor_in_airstream_fraction,
            fan_power_minimum_flow_rate_input_method=fan_power_minimum_flow_rate_input_method,
            fan_power_minimum_flow_rate_fraction=fan_power_minimum_flow_rate_fraction,
            fan_power_coefficient_1=fpc_1, fan_power_coefficient_2=fpc_2,
            fan_power_coefficient_3=fpc_3, fan_power_coefficient_4=fpc_4,
            fan_power_coefficient_5=fpc_5, end_use_subcategory=end_use_subcategory)
    elif fan_json['type'] == 'ZoneExhaust':
        fan = create_fan_zone_exhaust_from_json(
            model, fan_json, fan_name=fan_name, fan_efficiency=fan_efficiency,
            pressure_rise=pressure_rise,
            system_availability_manager_coupling_mode=system_availability_manager_coupling_mode,
            end_use_subcategory=end_use_subcategory)
    return fan


def create_fan_constant_volume_from_json(
        model, fan_json, fan_name=None, fan_efficiency=None, pressure_rise=None,
        motor_efficiency=None, motor_in_airstream_fraction=None, end_use_subcategory=None):
    """Creates a constant volume fan from a json."""
    fan_efficiency = fan_json['fan_efficiency'] \
        if fan_efficiency is None else fan_efficiency
    pressure_rise = fan_json['pressure_rise'] \
        if pressure_rise is None else pressure_rise
    motor_efficiency = fan_json['motor_efficiency'] \
        if motor_efficiency is None else motor_efficiency
    motor_in_airstream_fraction = fan_json['motor_in_airstream_fraction'] \
        if motor_in_airstream_fraction is None else motor_in_airstream_fraction

    pressure_rise = PRESSURE.to_unit([pressure_rise], 'Pa', 'inH2O')[0]

    fan = openstudio_model.FanConstantVolume(model)
    _apply_base_fan_variables(
        fan, fan_name, fan_efficiency, pressure_rise, end_use_subcategory)
    if motor_efficiency is not None:
        fan.setMotorEfficiency(motor_efficiency)
    if motor_in_airstream_fraction is not None:
        fan.setMotorInAirstreamFraction(motor_in_airstream_fraction)
    return fan


def create_fan_on_off_from_json(
        model, fan_json, fan_name=None, fan_efficiency=None, pressure_rise=None,
        motor_efficiency=None, motor_in_airstream_fraction=None, end_use_subcategory=None):
    """Creates a on off fan from a json."""
    fan_efficiency = fan_json['fan_efficiency'] \
        if fan_efficiency is None else fan_efficiency
    pressure_rise = fan_json['pressure_rise'] \
        if pressure_rise is None else pressure_rise
    motor_efficiency = fan_json['motor_efficiency'] \
        if motor_efficiency is None else motor_efficiency
    motor_in_airstream_fraction = fan_json['motor_in_airstream_fraction'] \
        if motor_in_airstream_fraction is None else motor_in_airstream_fraction

    pressure_rise = PRESSURE.to_unit([pressure_rise], 'Pa', 'inH2O')[0]

    fan = openstudio_model.FanOnOff(model)
    _apply_base_fan_variables(
        fan, fan_name, fan_efficiency, pressure_rise, end_use_subcategory)
    if motor_efficiency is not None:
        fan.setMotorEfficiency(motor_efficiency)
    if motor_in_airstream_fraction is not None:
        fan.setMotorInAirstreamFraction(motor_in_airstream_fraction)
    return fan


def create_fan_variable_volume_from_json(
        model, fan_json, fan_name=None, fan_efficiency=None, pressure_rise=None,
        motor_efficiency=None, motor_in_airstream_fraction=None,
        fan_power_minimum_flow_rate_input_method=None,
        fan_power_minimum_flow_rate_fraction=None,
        end_use_subcategory=None, fan_power_coefficient_1=None,
        fan_power_coefficient_2=None, fan_power_coefficient_3=None,
        fan_power_coefficient_4=None, fan_power_coefficient_5=None):
    """Creates a variable volume fan from a json."""
    fan_efficiency = fan_json['fan_efficiency'] \
        if fan_efficiency is None else fan_efficiency
    pressure_rise = fan_json['pressure_rise'] \
        if pressure_rise is None else pressure_rise
    motor_efficiency = fan_json['motor_efficiency'] \
        if motor_efficiency is None else motor_efficiency
    motor_in_airstream_fraction = fan_json['motor_in_airstream_fraction'] \
        if motor_in_airstream_fraction is None else motor_in_airstream_fraction
    if fan_power_minimum_flow_rate_input_method is None:
        fan_power_minimum_flow_rate_input_method = \
            fan_json['fan_power_minimum_flow_rate_input_method']
    if fan_power_minimum_flow_rate_fraction is None:
        fan_power_minimum_flow_rate_fraction = \
            fan_json['fan_power_minimum_flow_rate_fraction']

    pressure_rise = PRESSURE.to_unit([pressure_rise], 'Pa', 'inH2O')[0]

    fan = openstudio_model.FanVariableVolume(model)
    _apply_base_fan_variables(
        fan, fan_name, fan_efficiency, pressure_rise, end_use_subcategory)
    if motor_efficiency is not None:
        fan.setMotorEfficiency(motor_efficiency)
    if motor_in_airstream_fraction is not None:
        fan.setMotorInAirstreamFraction(motor_in_airstream_fraction)
    if fan_power_minimum_flow_rate_input_method is not None:
        fan.setFanPowerMinimumFlowRateInputMethod(fan_power_minimum_flow_rate_input_method)
    if fan_power_minimum_flow_rate_fraction is not None:
        fan.setFanPowerMinimumFlowFraction(fan_power_minimum_flow_rate_fraction)
    if fan_power_coefficient_1 is not None:
        fan.setFanPowerCoefficient1(fan_power_coefficient_1)
    if fan_power_coefficient_2 is not None:
        fan.setFanPowerCoefficient2(fan_power_coefficient_2)
    if fan_power_coefficient_3 is not None:
        fan.setFanPowerCoefficient3(fan_power_coefficient_3)
    if fan_power_coefficient_4 is not None:
        fan.setFanPowerCoefficient4(fan_power_coefficient_4)
    if fan_power_coefficient_5 is not None:
        fan.setFanPowerCoefficient5(fan_power_coefficient_5)
    return fan


def create_fan_zone_exhaust_from_json(
        model, fan_json, fan_name=None, fan_efficiency=None, pressure_rise=None,
        system_availability_manager_coupling_mode=None, end_use_subcategory=None):
    """Creates a FanZoneExhaust from a json."""
    fan_efficiency = fan_json['fan_efficiency'] \
        if fan_efficiency is None else fan_efficiency
    pressure_rise = fan_json['pressure_rise'] \
        if pressure_rise is None else pressure_rise

    pressure_rise = PRESSURE.to_unit([pressure_rise], 'Pa', 'inH2O')[0]

    fan = openstudio_model.FanZoneExhaust(model)
    _apply_base_fan_variables(
        fan, fan_name, fan_efficiency, pressure_rise, end_use_subcategory)
    if system_availability_manager_coupling_mode is not None:
        fan.setSystemAvailabilityManagerCouplingMode(
            system_availability_manager_coupling_mode)
    return fan


def _apply_base_fan_variables(fan, fan_name=None, fan_efficiency=None,
                              pressure_rise=None, end_use_subcategory=None):
    if fan_name is not None:
        fan.setName(fan_name)
    if fan_efficiency is not None:
        fan.setFanEfficiency(fan_efficiency)
    if pressure_rise is not None:
        fan.setPressureRise(pressure_rise)
    if end_use_subcategory is not None:
        fan.setEndUseSubcategory(end_use_subcategory)
    return fan


def fan_change_impeller_efficiency(fan, impeller_eff):
    """Changes the fan impeller efficiency and also the fan total efficiency.

    Motor efficiency is preserved.

    Args:
        fan: [OpenStudio::Model::StraightComponent] Fan object. Allowable types include
            FanConstantVolume, FanOnOff, FanVariableVolume, and FanZoneExhaust.
        impeller_eff: [Double] impeller efficiency (0.0 to 1.0).
    """
    # Get the existing motor efficiency
    existing_motor_eff = 0.7
    if fan.to_FanZoneExhaust().is_initialized():
        existing_motor_eff = fan.motorEfficiency()
    # Calculate the new total efficiency
    new_total_eff = existing_motor_eff * impeller_eff
    # Set the revised motor and total fan efficiencies
    fan.setFanEfficiency(new_total_eff)


def fan_baseline_impeller_efficiency(fan):
    """Assume that the fan efficiency is 65% for normal fans and 55% for small fans."""
    fan_impeller_eff = 0.65
    if is_small_fan(fan):
        fan_impeller_eff = 0.55
    return fan_impeller_eff


def is_small_fan(fan):
    """Zone exhaust fans, FCU fans, and VAV terminal fans all count as small fans.

    Small fans get different impeller efficiencies and motor efficiencies than
    other fans.

    Args:
        fan: [OpenStudio::Model::StraightComponent] Fan object. Allowable types
            includeFanConstantVolume, FanOnOff, FanVariableVolume, and FanZoneExhaust
    """
    is_small = False
    # Exhaust fan
    if fan.to_FanZoneExhaust().is_initialized():
        is_small = True
    # Fan coil unit, unit heater, PTAC, PTHP, VRF terminals, WSHP, ERV
    elif fan.containingZoneHVACComponent().is_initialized():
        zone_hvac = fan.containingZoneHVACComponent().get()
        if zone_hvac.to_ZoneHVACFourPipeFanCoil().is_initialized():
            is_small = True
        elif zone_hvac.to_ZoneHVACPackagedTerminalAirConditioner().is_initialized() or \
                zone_hvac.to_ZoneHVACPackagedTerminalHeatPump().is_initialized() or \
                zone_hvac.to_ZoneHVACTerminalUnitVariableRefrigerantFlow().is_initialized() or \
                zone_hvac.to_ZoneHVACWaterToAirHeatPump().is_initialized() or \
                zone_hvac.to_ZoneHVACEnergyRecoveryVentilator().is_initialized():
            is_small = True
    # Powered VAV terminal
    elif fan.containingHVACComponent().is_initialized():
        zone_hvac = fan.containingHVACComponent().get()
        if zone_hvac.to_AirTerminalSingleDuctParallelPIUReheat().is_initialized() or \
                zone_hvac.to_AirTerminalSingleDuctSeriesPIUReheat().is_initialized():
            is_small = True
    return is_small



# coding=utf-8
"""Module taken from OpenStudio-standards.

https://github.com/NREL/openstudio-standards/blob/master/
lib/openstudio-standards/prototypes/common/objects/Prototype.utilities.rb
"""
from __future__ import division

import re

from ladybug.datatype.power import Power

from honeybee_openstudio.openstudio import openstudio_model

POWER = Power()


def kw_per_ton_to_cop(kw_per_ton):
    """A helper method to convert from kW/ton to COP."""
    return 3.517 / kw_per_ton


def eer_to_cop_no_fan(eer, capacity_w=None):
    """Convert from EER to COP.

    If capacity is not supplied, use DOE Prototype Building method.
    If capacity is supplied, use the 90.1-2013 method.

    Args:
        eer: [Double] Energy Efficiency Ratio (EER).
        capacity_w: [Double] the heating capacity at AHRI rating conditions, in W.
    """
    if capacity_w is None:
        # From Thornton et al. 2011
        # r is the ratio of supply fan power to total equipment power at the rating condition,
        # assumed to be 0.12 for the reference buildings per Thornton et al. 2011.
        r = 0.12
        cop = ((eer / POWER.to_unit([1.0], 'Btu/h', 'W')[0]) + r) / (1 - r)
    else:
        # The 90.1-2013 method
        # Convert the capacity to Btu/hr
        capacity_btu_per_hr = POWER.to_unit([capacity_w], 'Btu/h', 'W')[0]
        cop = (7.84E-8 * eer * capacity_btu_per_hr) + (0.338 * eer)
    return cop


def hspf_to_cop_no_fan(hspf):
    """Convert from HSPF to COP (no fan) for heat pump heating coils.

    References - ASHRAE 90.1-2013. Appendix G.

    Args:
        hspf: [Double] heating seasonal performance factor (HSPF).
    """
    cop = (-0.0296 * hspf * hspf) + (0.7134 * hspf)
    return cop


def ems_friendly_name(name):
    """Converts existing string to ems friendly string."""
    # replace white space and special characters with underscore
    # \W is equivalent to [^a-zA-Z0-9_]
    new_name = re.sub('[^A-Za-z0-9]', '_', str(name))
    # prepend ems_ in case the name starts with a number
    new_name = 'ems_{}'.format(new_name)
    return new_name


def convert_curve_biquadratic(coeffs, ip_to_si=True):
    """Convert biquadratic curves that are a function of temperature.

    From IP (F) to SI (C) or vice-versa.  The curve is of the form
    z = C1 + C2*x + C3*x^2 + C4*y + C5*y^2 + C6*x*y
    where C1, C2, ... are the coefficients,
    x is the first independent variable (in F or C)
    y is the second independent variable (in F or C)
    and z is the resulting value
    """
    if ip_to_si:
        # Convert IP curves to SI curves
        si_coeffs = []
        si_coeffs.append((coeffs[0] + (32.0 * (coeffs[1] + coeffs[3])) +
                          (1024.0 * (coeffs[2] + coeffs[4] + coeffs[5]))))
        si_coeffs.append(((9.0 / 5.0 * coeffs[1]) +
                          (576.0 / 5.0 * coeffs[2]) + (288.0 / 5.0 * coeffs[5])))
        si_coeffs.append((81.0 / 25.0 * coeffs[2]))
        si_coeffs.append(((9.0 / 5.0 * coeffs[3]) +
                          (576.0 / 5.0 * coeffs[4]) + (288.0 / 5.0 * coeffs[5])))
        si_coeffs.append((81.0 / 25.0 * coeffs[4]))
        si_coeffs.append((81.0 / 25.0 * coeffs[5]))
        return si_coeffs
    else:
        # Convert SI curves to IP curves
        ip_coeffs = []
        ip_coeffs.append((coeffs[0] - (160.0 / 9.0 * (coeffs[1] + coeffs[3])) +
                          (25600.0 / 81.0 * (coeffs[2] + coeffs[4] + coeffs[5]))))
        ip_coeffs.append((5.0 / 9.0 * (coeffs[1] - (320.0 / 9.0 * coeffs[2]) -
                                       (160.0 / 9.0 * coeffs[5]))))
        ip_coeffs.append((25.0 / 81.0 * coeffs[2]))
        ip_coeffs.append((5.0 / 9.0 * (coeffs[3] - (320.0 / 9.0 * coeffs[4]) -
                                       (160.0 / 9.0 * coeffs[5]))))
        ip_coeffs.append((25.0 / 81.0 * coeffs[4]))
        ip_coeffs.append((25.0 / 81.0 * coeffs[5]))
        return ip_coeffs


def create_curve_biquadratic(
        model, coeffs, crv_name, min_x, max_x, min_y, max_y, min_out, max_out):
    """Create a biquadratic curve."""
    curve = openstudio_model.CurveBiquadratic(model)
    curve.setName(crv_name)
    curve.setCoefficient1Constant(coeffs[0])
    curve.setCoefficient2x(coeffs[1])
    curve.setCoefficient3xPOW2(coeffs[2])
    curve.setCoefficient4y(coeffs[3])
    curve.setCoefficient5yPOW2(coeffs[4])
    curve.setCoefficient6xTIMESY(coeffs[5])
    if min_x is None:
        curve.setMinimumValueofx(min_x)
    if max_x is not None:
        curve.setMaximumValueofx(max_x)
    if min_y is not None:
        curve.setMinimumValueofy(min_y)
    if max_y is not None:
        curve.setMaximumValueofy(max_y)
    if min_out is not None:
        curve.setMinimumCurveOutput(min_out)
    if max_out is not None:
        curve.setMaximumCurveOutput(max_out)
    return curve


def create_curve_quadratic(
        model, coeffs, crv_name, min_x, max_x, min_out, max_out, is_dimensionless=False):
    """Create a quadratic curve."""
    curve = openstudio_model.CurveQuadratic(model)
    curve.setName(crv_name)
    curve.setCoefficient1Constant(coeffs[0])
    curve.setCoefficient2x(coeffs[1])
    curve.setCoefficient3xPOW2(coeffs[2])
    if min_x is None:
        curve.setMinimumValueofx(min_x)
    if max_x is not None:
        curve.setMaximumValueofx(max_x)
    if min_out is not None:
        curve.setMinimumCurveOutput(min_out)
    if max_out is not None:
        curve.setMaximumCurveOutput(max_out)
    if is_dimensionless:
        curve.setInputUnitTypeforX('Dimensionless')
        curve.setOutputUnitType('Dimensionless')
    return curve


def rename_air_loop_nodes(model):
    """Renames air loop nodes to readable values."""
    # rename all hvac components on air loops
    for component in model.getHVACComponents():
        if component.to_Node().is_initialized():  # don't re-rename the node
            continue

        if component.airLoopHVAC().is_initialized():
            # rename water to air component outlet nodes
            if component.to_WaterToAirComponent().is_initialized():
                component = component.to_WaterToAirComponent().get()
                if component.airOutletModelObject().is_initialized():
                    component_outlet_object = component.airOutletModelObject().get()
                    if not component_outlet_object.to_Node().is_initialized():
                        continue
                    cp_name = component.nameString()
                    component_outlet_object.setName('{} Outlet Air Node'.format(cp_name))

            # rename air to air component nodes
            if component.to_AirToAirComponent().is_initialized():
                component = component.to_AirToAirComponent().get()
                if component.primaryAirOutletModelObject().is_initialized():
                    component_outlet_object = component.primaryAirOutletModelObject().get()
                    if not component_outlet_object.to_Node().is_initialized():
                        continue
                    cp_name = component.nameString()
                    component_outlet_object.setName(
                        '{} Primary Outlet Air Node'.format(cp_name))
                if component.secondaryAirInletModelObject().is_initialized():
                    component_inlet_object = component.secondaryAirInletModelObject().get()
                    if not component_inlet_object.to_Node().is_initialized():
                        continue
                    cp_name = component.nameString()
                    component_inlet_object.setName(
                        '{} Secondary Inlet Air Node'.format(cp_name))

            # rename straight component outlet nodes
            if component.to_StraightComponent().is_initialized():
                st_comp = component.to_StraightComponent().get()
                if st_comp.outletModelObject().is_initialized():
                    component_outlet_object = st_comp.outletModelObject().get()
                    if not component_outlet_object.to_Node().is_initialized():
                        continue
                    cp_name = component.nameString()
                    component_outlet_object.setName('{} Outlet Air Node'.format(cp_name))

        # rename zone hvac component nodes
        if component.to_ZoneHVACComponent().is_initialized():
            component = component.to_ZoneHVACComponent().get()
            if component.airInletModelObject().is_initialized():
                component_inlet_object = component.airInletModelObject().get()
                if not component_inlet_object.to_Node().is_initialized():
                    continue
                cp_name = component.nameString()
                component_inlet_object.setName('{} Inlet Air Node'.format(cp_name))

            if component.airOutletModelObject().is_initialized():
                component_outlet_object = component.airOutletModelObject().get()
                if not component_outlet_object.to_Node().is_initialized():
                    continue
                cp_name = component.nameString()
                component_outlet_object.setName('{} Outlet Air Node'.format(cp_name))

    # rename supply side nodes
    for air_loop in model.getAirLoopHVACs():
        air_loop_name = air_loop.nameString()
        air_loop.demandInletNode().setName('{} Demand Inlet Node'.format(air_loop_name))
        air_loop.demandOutletNode().setName('{} Demand Outlet Node'.format(air_loop_name))
        air_loop.supplyInletNode().setName('{} Supply Inlet Node'.format(air_loop_name))
        air_loop.supplyOutletNode().setName('{} Supply Outlet Node'.format(air_loop_name))

        if air_loop.reliefAirNode().is_initialized():
            relief_node = air_loop.reliefAirNode().get()
            relief_node.setName('{} Relief Air Node'.format(air_loop_name))

        if air_loop.mixedAirNode().is_initialized():
            mixed_node = air_loop.mixedAirNode().get()
            mixed_node.setName('{} Mixed Air Node'.format(air_loop_name))

        # rename outdoor air system and nodes
        if air_loop.airLoopHVACOutdoorAirSystem().is_initialized():
            oa_system = air_loop.airLoopHVACOutdoorAirSystem().get()
            if oa_system.outboardOANode().is_initialized():
                oa_node = oa_system.outboardOANode().get()
                oa_node.setName('{} Outdoor Air Node'.format(air_loop_name))

    # rename zone air and terminal nodes
    for zone in model.getThermalZones():
        zone_name = zone.nameString()
        zone.zoneAirNode().setName('{} Zone Air Node'.format(zone_name))
        if zone.returnAirModelObject().is_initialized():
            zone_node = zone.returnAirModelObject().get()
            zone_node.setName('{} Return Air Node'.format(zone_name))

        if zone.airLoopHVACTerminal().is_initialized():
            terminal_unit = zone.airLoopHVACTerminal().get()
            unit_name = terminal_unit.nameString()
            if terminal_unit.to_StraightComponent().is_initialized():
                component = terminal_unit.to_StraightComponent().get()
                if component.inletModelObject().is_initialized():
                    in_node = component.inletModelObject().get()
                    in_node.setName('{} Inlet Air Node'.format(unit_name))

    # rename zone equipment list objects
    for obj in model.getZoneHVACEquipmentLists():
        try:
            zone = obj.thermalZone()
            obj.setName('{} Zone HVAC Equipment List'.format(zone.nameString()))
        except Exception:
            obj.remove()  # missing thermal zone

    return model


def rename_plant_loop_nodes(model):
    """Renames plant loop nodes to readable values."""
    # rename all hvac components on plant loops
    for component in model.getHVACComponents():
        if component.to_Node().is_initialized():  # don't re-rename the node
            continue

        if component.plantLoop().is_initialized():
            # rename straight component nodes
            # some inlet or outlet nodes may get renamed again
            if component.to_StraightComponent().is_initialized():
                st_comp = component.to_StraightComponent().get()
                if st_comp.inletModelObject().is_initialized():
                    component_inlet_object = st_comp.inletModelObject().get()
                    if not component_inlet_object.to_Node().is_initialized():
                        continue
                    cp_name = component.nameString()
                    component_inlet_object.setName(
                        '{} Inlet Water Node'.format(cp_name))

                if st_comp.outletModelObject().is_initialized():
                    component_outlet_object = st_comp.outletModelObject().get()
                    if not component_outlet_object.to_Node().is_initialized():
                        continue
                    cp_name = component.nameString()
                    component_outlet_object.setName(
                        '{} Outlet Water Node'.format(cp_name))

            # rename water to air component nodes
            if component.to_WaterToAirComponent().is_initialized():
                component = component.to_WaterToAirComponent().get()
                if component.waterInletModelObject().is_initialized():
                    component_inlet_object = component.waterInletModelObject().get()
                    if not component_inlet_object.to_Node().is_initialized():
                        continue
                    cp_name = component.nameString()
                    component_inlet_object.setName(
                        '{} Inlet Water Node'.format(cp_name))

                if component.waterOutletModelObject().is_initialized():
                    component_outlet_object = component.waterOutletModelObject().get()
                    if not component_outlet_object.to_Node().is_initialized():
                        continue
                    cp_name = component.nameString()
                    component_outlet_object.setName(
                        '{} Outlet Water Node'.format(cp_name))

            # rename water to water component nodes
            if component.to_WaterToWaterComponent().is_initialized():
                component = component.to_WaterToWaterComponent().get()
                if component.demandInletModelObject().is_initialized():
                    demand_inlet_object = component.demandInletModelObject().get()
                    if not demand_inlet_object.to_Node().is_initialized():
                        continue
                    cp_name = component.nameString()
                    demand_inlet_object.setName(
                        '{} Demand Inlet Water Node'.format(cp_name))

                if component.demandOutletModelObject().is_initialized():
                    demand_outlet_object = component.demandOutletModelObject().get()
                    if not demand_outlet_object.to_Node().is_initialized():
                        continue
                    cp_name = component.nameString()
                    demand_outlet_object.setName(
                        '{} Demand Outlet Water Node'.format(cp_name))

                if component.supplyInletModelObject().is_initialized():
                    supply_inlet_object = component.supplyInletModelObject().get()
                    if not supply_inlet_object.to_Node().is_initialized():
                        continue
                    cp_name = component.nameString()
                    supply_inlet_object.setName(
                        '{} Supply Inlet Water Node'.format(cp_name))

                if component.supplyOutletModelObject().is_initialized():
                    supply_outlet_object = component.supplyOutletModelObject().get()
                    if not supply_outlet_object.to_Node().is_initialized():
                        continue
                    cp_name = component.nameString()
                    supply_outlet_object.setName(
                        '{} Supply Outlet Water Node'.format(cp_name))

    # rename plant nodes
    for plant_loop in model.getPlantLoops():
        pl_name = plant_loop.nameString()
        plant_loop.demandInletNode().setName('{} Demand Inlet Node'.format(pl_name))
        plant_loop.demandOutletNode().setName('{} Demand Outlet Node'.format(pl_name))
        plant_loop.supplyInletNode().setName('{} Supply Inlet Node'.format(pl_name))
        plant_loop.supplyOutletNode().setName('{} Supply Outlet Node'.format(pl_name))

    return model



# coding=utf-8
"""OpenStudio material translators."""
from __future__ import division

from honeybee_energy.material.opaque import EnergyMaterial, EnergyMaterialNoMass, \
    EnergyMaterialVegetation
from honeybee_energy.material.glazing import EnergyWindowMaterialGlazing, \
    EnergyWindowMaterialSimpleGlazSys
from honeybee_energy.material.gas import EnergyWindowMaterialGas, \
    EnergyWindowMaterialGasMixture, EnergyWindowMaterialGasCustom
from honeybee_energy.material.frame import EnergyWindowFrame
from honeybee_energy.material.shade import EnergyWindowMaterialShade, \
    EnergyWindowMaterialBlind

from honeybee_openstudio.openstudio import OSStandardOpaqueMaterial, \
    OSMasslessOpaqueMaterial, OSRoofVegetation, OSStandardGlazing, OSSimpleGlazing, \
    OSGas, OSGasMixture, OSShade, OSBlind, OSWindowPropertyFrameAndDivider


def opaque_material_to_openstudio(material, os_model):
    """Convert Honeybee EnergyMaterial to OpenStudio StandardOpaqueMaterial."""
    os_opaque_mat = OSStandardOpaqueMaterial(os_model)
    os_opaque_mat.setName(material.identifier)
    if material._display_name is not None:
        os_opaque_mat.setDisplayName(material.display_name)
    os_opaque_mat.setThickness(material.thickness)
    os_opaque_mat.setConductivity(material.conductivity)
    os_opaque_mat.setDensity(material.density)
    os_opaque_mat.setSpecificHeat(material.specific_heat)
    os_opaque_mat.setRoughness(material.roughness)
    os_opaque_mat.setThermalAbsorptance(material.thermal_absorptance)
    os_opaque_mat.setSolarAbsorptance(material.solar_absorptance)
    os_opaque_mat.setVisibleAbsorptance(material.visible_absorptance)
    return os_opaque_mat


def opaque_no_mass_material_to_openstudio(material, os_model):
    """Convert Honeybee EnergyMaterialNoMass to OpenStudio MasslessOpaqueMaterial."""
    os_nomass_mat = OSMasslessOpaqueMaterial(os_model)
    os_nomass_mat.setName(material.identifier)
    if material._display_name is not None:
        os_nomass_mat.setDisplayName(material.display_name)
    os_nomass_mat.setThermalResistance(material.r_value)
    os_nomass_mat.setRoughness(material.roughness)
    os_nomass_mat.setThermalAbsorptance(material.thermal_absorptance)
    os_nomass_mat.setSolarAbsorptance(material.solar_absorptance)
    os_nomass_mat.setVisibleAbsorptance(material.visible_absorptance)
    return os_nomass_mat


def vegetation_material_to_openstudio(material, os_model):
    """Convert Honeybee EnergyMaterialVegetation to OpenStudio RoofVegetation."""
    os_veg_mat = OSRoofVegetation(os_model)
    os_veg_mat.setName(material.identifier)
    if material._display_name is not None:
        os_veg_mat.setDisplayName(material.display_name)
    os_veg_mat.setThickness(material.thickness)
    os_veg_mat.setConductivityofDrySoil(material.conductivity)
    os_veg_mat.setDensityofDrySoil(material.density)
    os_veg_mat.setSpecificHeatofDrySoil(material.specific_heat)
    os_veg_mat.setRoughness(material.roughness)
    os_veg_mat.setThermalAbsorptance(material.soil_thermal_absorptance)
    os_veg_mat.setSolarAbsorptance(material.soil_solar_absorptance)
    os_veg_mat.setVisibleAbsorptance(material.soil_visible_absorptance)
    os_veg_mat.setHeightofPlants(material.plant_height)
    os_veg_mat.setLeafAreaIndex(material.leaf_area_index)
    os_veg_mat.setLeafReflectivity(material.leaf_reflectivity)
    os_veg_mat.setLeafEmissivity(material.leaf_emissivity)
    os_veg_mat.setMinimumStomatalResistance(material.min_stomatal_resist)
    os_veg_mat.setSaturationVolumetricMoistureContentoftheSoilLayer(material.sat_vol_moist_cont)
    os_veg_mat.setResidualVolumetricMoistureContentoftheSoilLayer(material.residual_vol_moist_cont)
    os_veg_mat.setInitialVolumetricMoistureContentoftheSoilLayer(material.init_vol_moist_cont)
    os_veg_mat.setMoistureDiffusionCalculationMethod(material.moist_diff_model)
    return os_veg_mat


def glazing_material_to_openstudio(material, os_model):
    """Convert Honeybee EnergyWindowMaterialGlazing to OpenStudio StandardGlazing."""
    os_glazing = OSStandardGlazing(os_model)
    os_glazing.setName(material.identifier)
    if material._display_name is not None:
        os_glazing.setDisplayName(material.display_name)
    os_glazing.setThickness(material.thickness)
    os_glazing.setSolarTransmittanceatNormalIncidence(material.solar_transmittance)
    os_glazing.setFrontSideSolarReflectanceatNormalIncidence(material.solar_reflectance)
    os_glazing.setBackSideSolarReflectanceatNormalIncidence(material.solar_reflectance_back)
    os_glazing.setVisibleTransmittanceatNormalIncidence(material.visible_transmittance)
    os_glazing.setFrontSideVisibleReflectanceatNormalIncidence(material.visible_reflectance)
    os_glazing.setBackSideVisibleReflectanceatNormalIncidence(material.visible_reflectance_back)
    os_glazing.setInfraredTransmittanceatNormalIncidence(material.infrared_transmittance)
    os_glazing.setFrontSideInfraredHemisphericalEmissivity(material.emissivity)
    os_glazing.setBackSideInfraredHemisphericalEmissivity(material.emissivity_back)
    os_glazing.setThermalConductivity(material.conductivity)
    os_glazing.setDirtCorrectionFactorforSolarandVisibleTransmittance(material.dirt_correction)
    os_glazing.setSolarDiffusing(material.solar_diffusing)
    return os_glazing


def simple_glazing_sys_material_to_openstudio(material, os_model):
    """Convert Honeybee EnergyWindowMaterialSimpleGlazSys to OpenStudio SimpleGlazing."""
    os_glz_sys = OSSimpleGlazing(os_model)
    os_glz_sys.setName(material.identifier)
    if material._display_name is not None:
        os_glz_sys.setDisplayName(material.display_name)
    os_glz_sys.setUFactor(material.u_factor)
    os_glz_sys.setSolarHeatGainCoefficient(material.shgc)
    os_glz_sys.setVisibleTransmittance(material.vt)
    return os_glz_sys


def gas_material_to_openstudio(material, os_model):
    """Convert Honeybee EnergyWindowMaterialGas to OpenStudio Gas."""
    os_gas = OSGas(os_model)
    os_gas.setName(material.identifier)
    if material._display_name is not None:
        os_gas.setDisplayName(material.display_name)
    os_gas.setThickness(material.thickness)
    os_gas.setGasType(material.gas_type)
    return os_gas


def gas_mixture_material_to_openstudio(material, os_model):
    """Convert Honeybee EnergyWindowMaterialGasMixture to OpenStudio GasMixture."""
    os_gas_mix = OSGasMixture(os_model)
    os_gas_mix.setName(material.identifier)
    if material._display_name is not None:
        os_gas_mix.setDisplayName(material.display_name)
    os_gas_mix.setThickness(material.thickness)
    for i in range(len(material.gas_types)):
        os_gas_mix.setGasType(i, material.gas_types[i])
        os_gas_mix.setGasFraction(i, material.gas_fractions[i])
    return os_gas_mix


def gas_custom_material_to_openstudio(material, os_model):
    """Convert Honeybee EnergyWindowMaterialGasCustom to OpenStudio Gas."""
    os_gas_custom = OSGas(os_model)
    os_gas_custom.setName(material.identifier)
    if material._display_name is not None:
        os_gas_custom.setDisplayName(material.display_name)
    os_gas_custom.setThickness(material.thickness)
    os_gas_custom.setGasType('Custom')
    os_gas_custom.setConductivityCoefficientA(material.conductivity_coeff_a)
    os_gas_custom.setViscosityCoefficientA(material.viscosity_coeff_a)
    os_gas_custom.setSpecificHeatCoefficientA(material.specific_heat_coeff_a)
    os_gas_custom.setConductivityCoefficientB(material.conductivity_coeff_b)
    os_gas_custom.setViscosityCoefficientB(material.viscosity_coeff_b)
    os_gas_custom.setSpecificHeatCoefficientB(material.specific_heat_coeff_b)
    os_gas_custom.setConductivityCoefficientC(material.conductivity_coeff_c)
    os_gas_custom.setViscosityCoefficientC(material.viscosity_coeff_c)
    os_gas_custom.setSpecificHeatCoefficientC(material.specific_heat_coeff_c)
    os_gas_custom.setSpecificHeatRatio(material.specific_heat_ratio)
    os_gas_custom.setMolecularWeight(material.molecular_weight)
    return os_gas_custom


def shade_material_to_openstudio(material, os_model):
    """Convert Honeybee EnergyWindowMaterialShade to OpenStudio Shade."""
    os_shade_mat = OSShade(os_model)
    os_shade_mat.setName(material.identifier)
    if material._display_name is not None:
        os_shade_mat.setDisplayName(material.display_name)
    os_shade_mat.setSolarTransmittance(material.solar_transmittance)
    os_shade_mat.setSolarReflectance(material.solar_reflectance)
    os_shade_mat.setVisibleTransmittance(material.visible_transmittance)
    os_shade_mat.setVisibleReflectance(material.visible_reflectance)
    os_shade_mat.setThermalHemisphericalEmissivity(material.emissivity)
    os_shade_mat.setThermalTransmittance(material.infrared_transmittance)
    os_shade_mat.setThickness(material.thickness)
    os_shade_mat.setConductivity(material.conductivity)
    os_shade_mat.setShadetoGlassDistance(material.distance_to_glass)
    os_shade_mat.setTopOpeningMultiplier(material.top_opening_multiplier)
    os_shade_mat.setBottomOpeningMultiplier(material.bottom_opening_multiplier)
    os_shade_mat.setLeftSideOpeningMultiplier(material.left_opening_multiplier)
    os_shade_mat.setRightSideOpeningMultiplier(material.right_opening_multiplier)
    os_shade_mat.setAirflowPermeability(material.airflow_permeability)
    return os_shade_mat


def blind_material_to_openstudio(material, os_model):
    """Convert Honeybee EnergyWindowMaterialBlind to OpenStudio Blind."""
    os_blind = OSBlind(os_model)
    os_blind.setName(material.identifier)
    if material._display_name is not None:
        os_blind.setDisplayName(material.display_name)
    os_blind.setSlatOrientation(material.slat_orientation)
    os_blind.setSlatWidth(material.slat_width)
    os_blind.setSlatSeparation(material.slat_separation)
    os_blind.setSlatThickness(material.slat_thickness)
    os_blind.setSlatAngle(material.slat_angle)
    os_blind.setSlatConductivity(material.slat_conductivity)
    os_blind.setSlatBeamSolarTransmittance(material.beam_solar_transmittance)
    os_blind.setFrontSideSlatBeamSolarReflectance(material.beam_solar_reflectance)
    os_blind.setBackSideSlatBeamSolarReflectance(material.beam_solar_reflectance_back)
    os_blind.setSlatDiffuseSolarTransmittance(material.diffuse_solar_transmittance)
    os_blind.setFrontSideSlatDiffuseSolarReflectance(material.diffuse_solar_reflectance)
    os_blind.setBackSideSlatDiffuseSolarReflectance(material.diffuse_solar_reflectance_back)
    os_blind.setSlatDiffuseVisibleTransmittance(material.diffuse_visible_transmittance)
    os_blind.setFrontSideSlatDiffuseVisibleReflectance(material.diffuse_visible_reflectance)
    os_blind.setBackSideSlatDiffuseVisibleReflectance(material.diffuse_visible_reflectance_back)
    os_blind.setSlatBeamVisibleTransmittance(material.beam_visible_transmittance)
    os_blind.setFrontSideSlatBeamVisibleReflectance(material.beam_visible_reflectance)
    os_blind.setBackSideSlatBeamVisibleReflectance(material.beam_visible_reflectance_back)
    os_blind.setSlatInfraredHemisphericalTransmittance(material.infrared_transmittance)
    os_blind.setFrontSideSlatInfraredHemisphericalEmissivity(material.emissivity)
    os_blind.setBackSideSlatInfraredHemisphericalEmissivity(material.emissivity_back)
    os_blind.setBlindtoGlassDistance(material.distance_to_glass)
    os_blind.setBlindTopOpeningMultiplier(material.top_opening_multiplier)
    os_blind.setBlindBottomOpeningMultiplier(material.bottom_opening_multiplier)
    os_blind.setBlindLeftSideOpeningMultiplier(material.left_opening_multiplier)
    os_blind.setBlindRightSideOpeningMultiplier(material.right_opening_multiplier)
    return os_blind


def frame_material_to_openstudio(material, os_model):
    """Convert Honeybee EnergyWindowFrame to OpenStudio WindowPropertyFrameAndDivider."""
    os_frame_mat = OSWindowPropertyFrameAndDivider(os_model)
    os_frame_mat.setName(material.identifier)
    if material._display_name is not None:
        os_frame_mat.setDisplayName(material.display_name)
    os_frame_mat.setFrameWidth(material.width)
    os_frame_mat.setFrameConductance(material.conductance)
    os_frame_mat.setRatioOfFrameEdgeGlassConductanceToCenterOfGlassConductance(
        material.edge_to_center_ratio)
    os_frame_mat.setFrameOutsideProjection(material.outside_projection)
    os_frame_mat.setFrameInsideProjection(material.inside_projection)
    os_frame_mat.setFrameThermalHemisphericalEmissivity(material.thermal_absorptance)
    os_frame_mat.setFrameSolarAbsorptance(material.solar_absorptance)
    os_frame_mat.setFrameVisibleAbsorptance(material.visible_absorptance)
    return os_frame_mat


def material_to_openstudio(material, os_model):
    """Convert any Honeybee energy material into an OpenStudio object.

    Args:
        material: A honeybee-energy Python object of a material layer.
        os_model: The OpenStudio Model object to which the Room will be added.

    Returns:
        An OpenStudio object for the material.
    """
    if isinstance(material, EnergyMaterial):
        return opaque_material_to_openstudio(material, os_model)
    elif isinstance(material, EnergyMaterialNoMass):
        return opaque_no_mass_material_to_openstudio(material, os_model)
    elif isinstance(material, EnergyMaterialVegetation):
        return vegetation_material_to_openstudio(material, os_model)
    elif isinstance(material, EnergyWindowMaterialGlazing):
        return glazing_material_to_openstudio(material, os_model)
    elif isinstance(material, EnergyWindowMaterialSimpleGlazSys):
        return simple_glazing_sys_material_to_openstudio(material, os_model)
    elif isinstance(material, EnergyWindowMaterialGas):
        return gas_material_to_openstudio(material, os_model)
    elif isinstance(material, EnergyWindowMaterialGasMixture):
        return gas_mixture_material_to_openstudio(material, os_model)
    elif isinstance(material, EnergyWindowMaterialGasCustom):
        return gas_custom_material_to_openstudio(material, os_model)
    elif isinstance(material, EnergyWindowFrame):
        return frame_material_to_openstudio(material, os_model)
    elif isinstance(material, EnergyWindowMaterialShade):
        return shade_material_to_openstudio(material, os_model)
    elif isinstance(material, EnergyWindowMaterialBlind):
        return blind_material_to_openstudio(material, os_model)
    else:
        raise ValueError(
            '{} is not a recognized Energy Material type'.format(type(material)))



# coding=utf-8
"""OpenStudio photovoltaic and generator translators."""
from __future__ import division

from honeybee_openstudio.openstudio import OSGeneratorPVWatts, \
    OSElectricLoadCenterDistribution, OSElectricLoadCenterInverterPVWatts


def pv_properties_to_openstudio(pv_properties, os_shade, os_model):
    """Convert Honeybee PVProperties to OpenStudio GeneratorPVWatts.

    Args:
        pv_properties: The Honeybee PVProperties object to be translated
            to OpenStudio.
        os_shade: The OpenStudio ShadingSurface object to which the GeneratorPVWatts
            is being assigned.
        os_model: The OpenStudio model to which the GeneratorPVWatts
            will be added.
    """
    # compte the system capacity
    rated_watts = pv_properties.rated_efficiency * 1000  # 1000W/m2 of solar irradiance
    sys_cap = int(os_shade.netArea() * pv_properties.active_area_fraction * rated_watts)
    # create the PVWatts generator and set all of the properties
    os_gen = OSGeneratorPVWatts(os_model, os_shade, sys_cap)
    os_gen.setName('{}..{}'.format(pv_properties.identifier, os_shade.nameString()))
    os_gen.setModuleType(pv_properties.module_type)
    os_gen.setArrayType(pv_properties.mounting_type)
    os_gen.setSystemLosses(pv_properties.system_loss_fraction)
    os_gen.setGroundCoverageRatio(pv_properties.tracking_ground_coverage_ratio)
    return os_gen


def electric_load_center_to_openstudio(load_center, os_gen_objects, os_model):
    """Convert Honeybee ElectricLoadCenter to OpenStudio ElectricLoadCenterDistribution.

    Args:
        load_center: The Honeybee ElectricLoadCenter object to be translated
            to OpenStudio.
        os_gen_objects: A list of the OpenStudio Generators objects that are
            controlled by the ElectricLoadCenter.
        os_model: The OpenStudio model to which the ElectricLoadCenterDistribution
            will be added.
    """
    # create the ElectricLoadCenter:Distribution and add the generators
    os_load_center = OSElectricLoadCenterDistribution(os_model)
    os_load_center.setName('Model Load Center Distribution')
    for os_gen in os_gen_objects:
        os_load_center.addGenerator(os_gen)
    os_load_center.setGeneratorOperationSchemeType('Baseload')
    os_load_center.setElectricalBussType('DirectCurrentWithInverter')
    # create the inverter and assign it
    inverter = OSElectricLoadCenterInverterPVWatts(os_model)
    inverter.setDCToACSizeRatio(load_center.inverter_dc_to_ac_size_ratio)
    inverter.setInverterEfficiency(load_center.inverter_efficiency)
    os_load_center.setInverter(inverter)
    return os_load_center



# coding=utf-8
"""OpenStudio ventilative cooling translators."""
from __future__ import division
import re

from honeybee.boundarycondition import Outdoors
from honeybee_energy.ventcool.control import VentilationControl

from honeybee_openstudio.openstudio import OSZoneVentilationWindandStackOpenArea, \
    OSZoneVentilationDesignFlowRate, OSAirflowNetworkCrack, OSAirflowNetworkSimpleOpening, \
    OSAirflowNetworkHorizontalOpening, OSAirflowNetworkReferenceCrackConditions, \
    OSOutputVariable, OSEnergyManagementSystemSensor, OSEnergyManagementSystemActuator, \
    OSEnergyManagementSystemProgram, OSEnergyManagementSystemProgramCallingManager


def ventilation_opening_to_openstudio(opening, os_model):
    """Convert VentilationOpening to OpenStudio ZoneVentilationWindandStackOpenArea.

    Args:
        opening: The Honeybee VentilationOpening object to be translated
            to OpenStudio. Note that this object must be assigned to a parent
            Aperture with a parent Room in order to be successfully translated.
        os_model: The OpenStudio model to which the ZoneVentilationWindandStackOpenArea
            will be added.
    """
    # check that a parent is assigned
    assert opening.parent is not None, 'VentilationOpening must be assigned ' \
        'to an Aperture or Door to translate to_openstudio.'
    # get the VentilationControl object from the room
    control = None
    room = None
    if opening.parent.has_parent:
        if opening.parent.parent.has_parent:
            room = opening.parent.parent.parent
            if room.properties.energy.window_vent_control is not None:
                control = room.properties.energy.window_vent_control
    if control is None:  # use default ventilation control
        control = VentilationControl()
    assert room is not None, 'VentilationOpening must have a parent Room to ' \
        'translate to_openstudio.'
    # process the properties on this object into IDF format
    angle = opening.parent.horizontal_orientation() \
        if opening.parent.normal.z != 1 else 0
    angle = angle % 360
    height = (opening.parent.geometry.max.z - opening.parent.geometry.min.z) * \
        opening.fraction_height_operable
    # create wind and stack object and set all of its properties
    os_opening = OSZoneVentilationWindandStackOpenArea(os_model)
    os_opening.setName('{}_Opening'.format(opening.parent.identifier))
    os_opening.setOpeningArea(opening.parent.area * opening.fraction_area_operable)
    os_opening.setHeightDifference(height)
    os_opening.setEffectiveAngle(angle)
    os_opening.setDischargeCoefficientforOpening(opening.discharge_coefficient)
    if opening.wind_cross_vent:
        os_opening.autocalculateOpeningEffectiveness()
    else:
        os_opening.setOpeningEffectiveness(0)
    # set the properties of the ventilation control
    os_opening.setMinimumIndoorTemperature(control.min_indoor_temperature)
    os_opening.setMaximumIndoorTemperature(control.max_indoor_temperature)
    os_opening.setMinimumOutdoorTemperature(control.min_outdoor_temperature)
    os_opening.setMaximumOutdoorTemperature(control.max_outdoor_temperature)
    os_opening.setDeltaTemperature(control.delta_temperature)
    if control.schedule.identifier != 'Always On':
        vent_sch = os_model.getScheduleByName(control.schedule.identifier)
        if vent_sch.is_initialized():
            os_vent_sch = vent_sch.get()
            os_opening.setOpeningAreaFractionSchedule(os_vent_sch)
    return os_opening


def ventilation_fan_to_openstudio(fan, os_model):
    """Convert VentilationFan to OpenStudio ZoneVentilationDesignFlowRate."""
    # create zone ventilation object and set identifier
    os_fan = OSZoneVentilationDesignFlowRate(os_model)
    os_fan.setName(fan.identifier)
    if fan._display_name is not None:
        os_fan.setDisplayName(fan.display_name)
    # assign fan properties
    os_fan.setDesignFlowRate(fan.flow_rate)
    os_fan.setFanPressureRise(fan.pressure_rise)
    os_fan.setFanTotalEfficiency(fan.efficiency)
    os_fan.setVentilationType(fan.ventilation_type)
    # set all of the ventilation control properties
    os_fan.setMinimumIndoorTemperature(fan.control.min_indoor_temperature)
    os_fan.setMaximumIndoorTemperature(fan.control.max_indoor_temperature)
    os_fan.setMinimumOutdoorTemperature(fan.control.min_outdoor_temperature)
    os_fan.setMaximumOutdoorTemperature(fan.control.max_outdoor_temperature)
    os_fan.setDeltaTemperature(fan.control.delta_temperature)
    # assign schedule if it exists
    if fan.control.schedule.identifier != 'Always On':
        vent_sch = os_model.getScheduleByName(fan.control.schedule.identifier)
        if vent_sch.is_initialized():
            os_vent_sch = vent_sch.get()
            os_fan.setSchedule(os_vent_sch)
    return os_fan


def ventilation_sim_control_to_openstudio(vent_sim_control, os_model):
    """Convert VentilationSimulationControl to OpenStudio.

    This method returns an AirflowNetworkReferenceCrackConditions that can
    be used for the rest of the AFN setup.
    """
    # create the AirflowNetworkSimulationControl object
    os_v_sim_ctrl = os_model.getAirflowNetworkSimulationControl()
    os_v_sim_ctrl.setName('Window Based Ventilative Cooling')
    os_v_sim_ctrl.setAirflowNetworkControl(vent_sim_control.vent_control_type)
    os_v_sim_ctrl.setBuildingType(vent_sim_control.building_type)
    os_v_sim_ctrl.setAzimuthAngleofLongAxisofBuilding(vent_sim_control.long_axis_angle)
    os_v_sim_ctrl.setBuildingAspectRatio(vent_sim_control.aspect_ratio)
    # create the AirflowNetworkReferenceCrackConditions that other cracks reference
    os_ref_crack = OSAirflowNetworkReferenceCrackConditions(os_model)
    os_ref_crack.setName('Reference Crack Conditions')
    os_ref_crack.setTemperature(vent_sim_control.reference_temperature)
    os_ref_crack.setBarometricPressure(vent_sim_control.reference_pressure)
    os_ref_crack.setHumidityRatio(vent_sim_control.reference_humidity_ratio)
    return os_ref_crack


def afn_crack_to_openstudio(afn_crack, os_model, os_reference_crack=None):
    """Convert Honeybee AFNCrack to OpenStudio AirflowNetworkCrack.

    Args:
        opening: The Honeybee VentilationOpening object to be translated
            to OpenStudio. Note that this object must be assigned to a parent
            Aperture with a parent Room in order to be successfully translated.
        os_model: The OpenStudio model to which the AirflowNetworkSurface
            will be added.
        os_reference_crack: An optional AirflowNetworkReferenceCrackConditions
            object to set the reference. If None, a default reference crack will
            be created. (Default: None).
    """
    flow_coefficient = afn_crack.flow_coefficient \
        if afn_crack.flow_coefficient > 1.0e-09 else 1.0e-09
    flow_exponent = afn_crack.flow_exponent
    if os_reference_crack is None:
        os_reference_crack = _default_reference_crack(os_model)
    os_crack = OSAirflowNetworkCrack(
        os_model, flow_coefficient, flow_exponent, os_reference_crack)
    return os_crack


def ventilation_opening_to_openstudio_afn(opening, os_model, os_reference_crack=None):
    """Convert Honeybee VentilationOpening to OpenStudio AirflowNetworkSimpleOpening.

    The returned output may also be a AirflowNetworkHorizontalOpening or a
    AirflowNetworkCrack if the opening is assigned to a parent Aperture or Door
    that is horizontal and so it cannot be represented with AirflowNetworkSimpleOpening.

    Args:
        opening: The Honeybee VentilationOpening object to be translated
            to OpenStudio. Note that this object must be assigned to a parent
            Aperture with a parent Room in order to be successfully translated.
        os_model: The OpenStudio model to which the AirflowNetworkSimpleOpening
            will be added.
        os_reference_crack: An optional AirflowNetworkReferenceCrackConditions
            object to set the reference when the ventilation opening is being
            translated to a large crack. This happens when the ventilation
            opening is horizontal and in an outdoor Face. If None, a default
            reference crack will be created. (Default: None).

    Returns:
        A tuple with two elements.

        -   os_opening -- The OpenStudio AFN SimpleOpening, HorizontalOpening
            or Crack that represents the ventilation opening.

        -   opening_factor - A number for the opening factor to be assigned to
            the parent OpenStudio AirflowNetworkSurface and incorporated into
            the EMS program.
    """
    # check that a parent is assigned
    assert opening.parent is not None, 'VentilationOpening must be assigned ' \
        'to an Aperture or Door to translate to_openstudio.'
    # get the tilt and BC of the parent so that we can use the correct AFN object
    srf_tilt = opening.parent.tilt
    srf_bc = opening.parent.boundary_condition
    # process the flow coefficient, flow exponent and fraction area operable
    flow_coeff = opening.flow_coefficient_closed \
        if opening.flow_coefficient_closed > 1.0e-09 else 1.0e-09
    flow_exponent = opening.flow_exponent_closed
    discharge_coeff = opening.discharge_coefficient
    two_way_thresh = opening.two_way_threshold
    opening_factor = opening.fraction_area_operable
    # create an opening obj
    if srf_tilt < 10 or srf_tilt > 170:
        if isinstance(srf_bc, Outdoors):
            # create a crack to represent an exterior in-operable horizontal skylight
            opening_factor = None
            if os_reference_crack is None:
                os_reference_crack = _default_reference_crack(os_model)
            os_opening = OSAirflowNetworkCrack(
                os_model, flow_coeff, flow_exponent, os_reference_crack)
        else:
            # create a HorizontalOpening object to for the interior horizontal window
            slope_ang = 90 - srf_tilt if srf_tilt < 10 else 90 - (180 - srf_tilt)
            os_opening = OSAirflowNetworkHorizontalOpening(
                os_model, flow_coeff, flow_exponent, slope_ang, discharge_coeff)
    else:
        # create the simple opening object for the Aperture or Door using default values
        os_opening = OSAirflowNetworkSimpleOpening(
          os_model, flow_coeff, flow_exponent, two_way_thresh, discharge_coeff)
    os_opening.setName('{}_Opening'.format(opening.parent.identifier))
    return os_opening, opening_factor


def ventilation_control_to_openstudio_afn(
        control, open_factors, os_sub_faces, os_zone_air_temp, os_model, room_id=''):
    """Convert Honeybee VentilationControl to OpenStudio EnergyManagementSystemProgram.

    Args:
        control: The Honeybee VentilationControl object to be translated to OpenStudio.
        open_factors: A list of numbers for the opening factor of each Subface
            to be controlled by the VentilationControl.
        os_sub_faces: A list of OpenStudio SubSurface objects that have AFN
            SimpleOpening or HorizontalOpening objects to be controlled by the EMS.
        os_zone_air_temp: The OpenStudio EnergyManagementSystemSensor object for the
            Zone Air Temperature that corresponds with the VentilationControl.
            If this sensor does not yet exist in the model, the zone_temperature_sensor
            function in this module can be used to create it.
        os_model: The OpenStudio model to which the EnergyManagementSystemProgram
            will be added.
        room_id: An optional Room identifier to be used to ensure the names used
            in the resulting EnergyManagementSystemProgram are unique to the
            Room to which the VentilationControl is applied.
    """
    # set up a schedule sensor if there's a schedule specified
    sch_sen_id = None
    if control.schedule.identifier != 'Always On':
        vent_sch = os_model.getScheduleByName(control.schedule.identifier)
        if vent_sch.is_initialized():
            sch_var = OSOutputVariable('Schedule Value', os_model)
            sch_var.setReportingFrequency('Timestep')
            sch_var.setKeyValue(control.schedule.identifier)
            sch_sensor = OSEnergyManagementSystemSensor(os_model, sch_var)
            sch_sen_id = 'SensorSch{}'.format(re.sub('[^A-Za-z0-9]', '', room_id))
            sch_sensor.setName(sch_sen_id)

    # create the actuators for each of the operable windows
    actuator_ids = []
    for os_sub_f in os_sub_faces:
        window_actuator = OSEnergyManagementSystemActuator(
            os_sub_f, 'AirFlow Network Window/Door Opening',
            'Venting Opening Factor')
        act_id = 'OpenFactor{}'.format(re.sub('[^A-Za-z0-9]', '', os_sub_f.nameString()))
        window_actuator.setName(act_id)
        actuator_ids.append(act_id)

    # create the first part of the EMS Program to open windows according to control logic
    logic_statements = []
    in_sen_id = os_zone_air_temp.nameString()
    min_in = control.min_indoor_temperature
    max_in = control.max_indoor_temperature
    min_out = control.min_outdoor_temperature
    max_out = control.max_outdoor_temperature
    d_in_out = control.delta_temperature
    if min_in != -100:
        logic_statements.append('({} > {})'.format(in_sen_id, min_in))
    if max_in != 100:
        logic_statements.append('({} < {})'.format(in_sen_id, max_in))
    if min_out != -100:
        logic_statements.append('(Outdoor_Sensor > {})'.format(min_out))
    if max_out != 100:
        logic_statements.append('(Outdoor_Sensor < {})'.format(max_out))
    if d_in_out != -100:
        logic_statements.append('(({} - Outdoor_Sensor) > {})'.format(in_sen_id, d_in_out))
    if sch_sen_id is not None:
        logic_statements.append('({} > 0)'.format(sch_sen_id))
    if len(logic_statements) == 0:  # no logic has been provided; always open windows
        complete_logic = 'IF (Outdoor_Sensor < 100)'
    else:
        complete_logic = 'IF {}'.format(' && '.join(logic_statements))

    # initialize the program and add the complete logic
    ems_program = OSEnergyManagementSystemProgram(os_model)
    prog_name = 'WindowOpening{}'.format(re.sub('[^A-Za-z0-9]', '', room_id))
    ems_program.setName(prog_name)
    ems_program.addLine(complete_logic)

    # loop through each of the actuators and open/close each window
    for act_id, open_factor in zip(actuator_ids, open_factors):
        ems_program.addLine('SET {} = {}'.format(act_id, open_factor))
    ems_program.addLine('ELSE')
    for act_id in actuator_ids:
        ems_program.addLine('SET {} = 0'.format(act_id))
    ems_program.addLine('ENDIF')
    return ems_program


def zone_temperature_sensor(os_zone, os_model):
    """Create an EnergyManagementSystemSensor for Zone Air Temperature."""
    os_zone_name = os_zone.nameString()
    in_var = OSOutputVariable('Zone Air Temperature', os_model)
    in_var.setReportingFrequency('Timestep')
    in_var.setKeyValue(os_zone_name)
    in_air_sensor = OSEnergyManagementSystemSensor(os_model, in_var)
    sensor_id = 'Sensor{}'.format(re.sub('[^A-Za-z0-9]', '', os_zone_name))
    in_air_sensor.setName(sensor_id)
    return in_air_sensor


def outdoor_temperature_sensor(os_model):
    """Create an EnergyManagementSystemSensor for Site Outdoor Air DryBulb Temperature.
    """
    out_var = OSOutputVariable('Site Outdoor Air Drybulb Temperature', os_model)
    out_var.setReportingFrequency('Timestep')
    out_var.setKeyValue('Environment')
    out_air_sensor = OSEnergyManagementSystemSensor(os_model, out_var)
    out_air_sensor.setName('Outdoor_Sensor')
    return out_air_sensor


def ventilation_control_program_manager(os_model):
    """Create an EMS Program Manager for all window opening."""
    os_prog_manager = OSEnergyManagementSystemProgramCallingManager(os_model)
    os_prog_manager.setName('Temperature_Controlled_Window_Opening')
    os_prog_manager.setCallingPoint('BeginTimestepBeforePredictor')
    return os_prog_manager


def _default_reference_crack(os_model):
    """Create an AFN Reference Crack with default characteristics."""
    default_name = 'Reference Crack Conditions'
    os_ref_crack = os_model.getAirflowNetworkReferenceCrackConditionsByName(default_name)
    if os_ref_crack.is_initialized():
        return os_ref_crack.get()
    os_ref_crack = OSAirflowNetworkReferenceCrackConditions(os_model)
    os_ref_crack.setTemperature(20)
    os_ref_crack.setBarometricPressure(101325)
    os_ref_crack.setHumidityRatio(0)
    return os_ref_crack
